
// Line (lumped RLC serial circuit model) transient analysis
// Step response
// [RLC] * N segments, voltage source with resistance Rin

// Line:
//        ___
//    o--|___|--uuu--*--  ...  --o
//         R1    L1  |
//  u_0             === C1       u_N
//                   |
//    o--------------*--  ...  --o
//         segment1
//

// Method: Runge-Kutta 4 (not symplectic, accuracy problem)
// Needs linking with math library (GCC option -lm)
// WARNING: unstable numerical solution possible
// Copyright: (c) Petr Peringer, FIT TU Brno, 2017

#include <stdio.h>
#include <stdlib.h>             // strtod()
#include <math.h>               // sqrt(), needs gcc -lm

// System: N * RLC series
#define LLEN 1.0                // line length [m]
#define N    100                // segment count
#define len  (LLEN/N)           // segment length [m]
#define SIZE 2*N                // state vector size of the system

// Parameters for 1m line segment (telephone line at 1kHz?)
const double R1 = len * 172.24e-3;      // 172.24 Ohm/km
const double C1 = len * 51.57e-12;      // 51.57 nF/km
const double L1 = len * 0.6129e-6;      // 0.6129 mH/km
const double uin = 1;           // 1V step
double Rin;                     // source resistance [Ohm]
double Rz;                      // load resistance [Ohm]

// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    for (int k = 0; k < SIZE; k += 2) {
        // 2 state variables:
        double i = y[k + 1];    // current in loop
        double u_C = y[k + 0];  // capacitor voltage
        // temporary variables:
        double i1 = (k == SIZE - 2) ? u_C / Rz : y[k + 3];      // next i
        double uprev = (k == 0) ? uin - Rin * i : y[k - 2];     // prev u
        // series RLC equations converted into C language assignments:
        double u_R = R1 * i;    // Ohm's law
        double u_L = uprev - u_C - u_R;
        yin[k + 0] = (i - i1) / C1;     // der(u_C_x)
        yin[k + 1] = u_L / L1;          // der(i_x)
    }
}

// Runge-Kutta-4 step
void RK4step(double t, double h, double *y);

// Experiment control
int main(int argc, char * argv[])
{
    const double TEND = 1e-7 / 3.5;
    double y[SIZE] = { 0, 0, }; // zero initial state
    const double Z = sqrt(L1 / C1);     // approximation (R1 omitted)
    // default:
    Rin = Z;                    // source resistance Z
    Rz = Z;                     // load resistance Z
    // arguments: generator resistance, load resistance
    if(argc>1) {
        Rin = strtod(argv[1],NULL);
        if(Rin<1) Rin = 1;      // limit
        if(Rin>1e4) Rin = 1e4;
        if(argc>2)
            Rz = strtod(argv[2],NULL);
        if(Rz<1) Rz = 1;        // limit
        if(Rz>1e6) Rz = 1e6;
    }

    double stepsize = TEND / 1e5;
    const unsigned NS = (TEND / stepsize) / 2000;       // delta

    printf("# N=%d L=%g len=%g\n", N, LLEN, len);
    printf("# R1=%g L1=%g C1=%g Z=%.4g\n", R1, L1, C1, Z);
    printf("# Rg=%.4g Rl=%.4g\n", Rin, Rz);
    unsigned steps = 0;
    double t = 0;
    while (t < TEND) {
        double u_C = y[0];
        double i = y[1];
        double u0 = uin - Rin * i;
        double u_R = R1 * i;
        double u_L = u0 - u_C - u_R;
        double uN = y[SIZE - 2 + 0];
        double iN = y[SIZE - 2 + 1];
        if (steps % NS == 0)
            printf("%10.6f % -11g % -11g  % -11g\n", t*1e9, u0, uN, i);
        RK4step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
        ++steps;
    }
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}
